import tensorflow as tf
from transformer_model.decoder import DecoderLayer
from transformer_model.positional_encoder import PositionalEncoder


class Decoder(tf.keras.layers.Layer):

    def __init__(self, num_layers, d_model, num_heads, dff, target_vocab_size, rate=0.1):
        """
        :param num_layers: Number of DecoderLayers to use.
        :param d_model: Dimensionality of the model
        :param num_heads: Number of heads to use in the Multiheaded Attention
        :param dff: input nodes of the Point-wise Feed Forward NN
        :param target_vocab_size: Number of unique words in the target set
        :param rate: Dropout rate
        """
        super(Decoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoder = PositionalEncoder(d_model=d_model)
        self.pos_encoding = self.pos_encoder.positional_encoding(target_vocab_size, d_model)

        self.dec_layers = [DecoderLayer(d_model, num_heads, dff, rate) for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, x, enc_output, training, look_ahead_mask, padding_mask):
        seq_len = tf.shape(x)[1]
        attention_weights = {}

        x = self.embedding(x)  # (batch_size, target_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](x, enc_output, training, look_ahead_mask, padding_mask)

            attention_weights['decoder_layer{}_block1'.format(i + 1)] = block1
            attention_weights['decoder_layer{}_block2'.format(i + 1)] = block2

        # x.shape == (batch_size, target_seq_len, d_model)
        return x, attention_weights

    


import time
from transformer_model.masker import Masker
import tensorflow as tf


class Trainer(object):
    def __init__(self, transformer, learning_rate, optimizer, epochs, train_dataset, test_dataset,
                 load_ckpt=True,
                 loss_object=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction='none'),
                 train_loss=tf.keras.metrics.Mean(name='train_loss'),
                 train_accuracy=tf.keras.metrics.SparseCategoricalCrossentropy(name='train_accuracy'),
                 checkpoint_path='./models/checkpoints/',
                 max_to_keep=5
                 ):
        """
        :param transformer: Instance of Transformer Class
        :param learning_rate: Learning Rate schedule
        :param optimizer: TF optimizer object
        :param epochs: Number of Epochs to train on
        :param train_dataset: training dataset
        :param test_dataset: testing dataset
        :param load_ckpt: bool to load recent checkpoint if True
        :param loss_object: loss calculation object
        :param train_loss: training loss object
        :param train_accuracy: training accuracy object
        :param checkpoint_path: the location of where the checkpoints are stored
        :param max_to_keep: maximum number of checkpoints to keep at a time
        """
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset

        self.loss_object = loss_object
        self.train_loss = train_loss
        self.train_accuracy = train_accuracy
        self.checkpoint_path = checkpoint_path
        self.mskr = Masker()

        self.transformer = transformer
        self.learning_rate = learning_rate
        self.optimizer = optimizer

        self.epochs = epochs
        self.ckpt = tf.train.Checkpoint(transformer=transformer, optimizer=optimizer)
        self.max_to_keep = max_to_keep
        self.ckpt_manager = tf.train.CheckpointManager(self.ckpt, self.checkpoint_path, self.max_to_keep)

        if load_ckpt:
            print('Attempting to load latest checkpoint...')
            if self.ckpt_manager.latest_checkpoint:
                self.ckpt.restore(self.ckpt_manager.latest_checkpoint)
                print('Latest Checkpoint Restored!!!')
            else:
                print(f'No checkpoint found...')
                print('Training from scratch')
                pass
        else:
            pass

    def loss_function(self, real, pred):
        """
        :param real: the actual output
        :param pred: the model output
        :return: the training loss
        """

        mask = tf.math.logical_not(tf.math.equal(real, 0))
        loss_ = self.loss_object(real, pred)

        mask = tf.cast(mask, dtype=loss_.dtype)
        loss_ *= mask

        return tf.reduce_mean(loss_)

    def create_masks(self, inp, tar):
        # Encoder padding mask
        enc_padding_mask = self.mskr.create_padding_mask(inp)

        # Used in the 2nd attention block in the decoder.
        # This padding mask is used to mask the encoder outputs.
        dec_padding_mask = self.mskr.create_padding_mask(inp)

        # Used in the 1st attention block in the decoder.
        # It is used to pad and mask future tokens in the input received by
        # the decoder.
        look_ahead_mask = self.mskr.create_look_ahead_mask(tf.shape(tar)[1])
        dec_target_padding_mask = self.mskr.create_padding_mask(tar)
        combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

        return enc_padding_mask, combined_mask, dec_padding_mask

    @tf.function(input_signature=[
        tf.TensorSpec(shape=(None, None), dtype=tf.int64),
        tf.TensorSpec(shape=(None, None), dtype=tf.int64),
    ])
    def train_step(self, inp, tar):
        tar_inp = tar[:, :-1]
        tar_real = tar[:, 1:]

        enc_padding_mask, combined_mask, dec_padding_mask = self.mskr.create_masks(inp, tar_inp)

        with tf.GradientTape() as tape:
            predictions, _ = self.transformer(inp, tar_inp,
                                              True,
                                              enc_padding_mask,
                                              combined_mask,
                                              dec_padding_mask)
            loss = self.loss_function(tar_real, predictions)

        gradients = tape.gradient(loss, self.transformer.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.transformer.trainable_variables))

        self.train_loss(loss)
        self.train_accuracy(tar_real, predictions)

    def train(self):
        loss_hist = {}
        acc_hist = {}
        for epoch in range(self.epochs):
            start = time.time()

            self.train_loss.reset_states()
            self.train_accuracy.reset_states()

            # inp -> client, tar -> qltm
            for (batch, (inp, tar)) in enumerate(self.train_dataset):
                self.train_step(inp, tar)

                if batch % 50 == 0:
                    print('Epoch {} Batch {} Loss {:.4f} Accuracy {:.4f}'.format(
                        epoch + 1, batch, self.train_loss.result(), self.train_accuracy.result()))

            if (epoch + 1) % 5 == 0:
                ckpt_save_path = self.ckpt_manager.save()
                print('Saving checkpoint for epoch {} at {}'.format(epoch + 1,
                                                                    ckpt_save_path))

            print('Epoch {} Loss {:.4f} Accuracy {:.4f}'.format(epoch + 1,
                                                                self.train_loss.result(),
                                                                self.train_accuracy.result()))
            loss_hist[epoch] = self.train_loss.result()
            acc_hist[epoch] = self.train_accuracy.result()

            print('Time taken for 1 epoch: {} secs\n'.format(time.time() - start))

        return loss_hist, acc_hist



import tensorflow as tf
import matplotlib.pyplot as plt


class Evaluator(object):
    def __init__(self, tokenizer_feat, tokenizer_tar, max_length, mskr, transformer):
        self.tokenizer_feat = tokenizer_feat
        self.tokenizer_tar = tokenizer_tar
        self.max_length = max_length
        self.mskr = mskr
        self.transformer = transformer

    def evaluate(self, inp_sentence):
        """
        This allows the ability to score the models effectiveness at predicting each output sequence
        :return: The model output as well as the attention weights from the calculation
        """

        start_token = [self.tokenizer_feat.vocab_size]
        end_token = [self.tokenizer_feat.vocab_size + 1]

        # inp sentence is portuguese, hence adding the start and end token
        inp_sentence = start_token + self.tokenizer_feat.encode(inp_sentence) + end_token
        encoder_input = tf.expand_dims(inp_sentence, 0)

        # as the target is english, the first word to the transformer should be the
        # english start token.
        decoder_input = [self.tokenizer_tar.vocab_size]
        output = tf.expand_dims(decoder_input, 0)

        for i in range(self.max_length):
            enc_padding_mask, combined_mask, dec_padding_mask = self.mskr.create_masks(
                encoder_input, output)

            # predictions.shape == (batch_size, seq_len, vocab_size)
            predictions, attention_weights = self.transformer(encoder_input,
                                                              output,
                                                              False,
                                                              enc_padding_mask,
                                                              combined_mask,
                                                              dec_padding_mask)

            # select the last word from the seq_len dimension
            predictions = predictions[:, -1:, :]  # (batch_size, 1, vocab_size)

            predicted_id = tf.cast(tf.argmax(predictions, axis=-1), tf.int32)

            # return the result if the predicted_id is equal to the end token
            if tf.equal(predicted_id, self.tokenizer_tar.vocab_size + 1):
                return tf.squeeze(output, axis=0), attention_weights

            # concatenate the predicted_id to the output which is given to the decoder
            # as its input.
            output = tf.concat([output, predicted_id], axis=-1)

        return tf.squeeze(output, axis=0), attention_weights

    def plot_attention_weights(self, attention, sentence, result, layer):
        """
        A handy function to plot the attention weights.
        """

        fig = plt.figure(figsize=(16, 8))

        sentence = self.tokenizer_feat.encode(sentence)

        attention = tf.squeeze(attention[layer], axis=0)

        for head in range(attention.shape[0]):
            ax = fig.add_subplot(2, 4, head + 1)

            # plot the attention weights
            ax.matshow(attention[head][:-1, :], cmap='viridis')

            fontdict = {'fontsize': 10}

            ax.set_xticks(range(len(sentence) + 2))
            ax.set_yticks(range(len(result)))

            ax.set_ylim(len(result) - 1.5, -0.5)

            ax.set_xticklabels(
                ['<start>'] + [self.tokenizer_feat.decode([i]) for i in sentence] + ['<end>'],
                fontdict=fontdict, rotation=90)

            ax.set_yticklabels([self.tokenizer_tar.decode([i]) for i in result
                                if i < self.tokenizer_tar.vocab_size],
                               fontdict=fontdict)

            ax.set_xlabel('Head {}'.format(head + 1))

        plt.tight_layout()
        plt.show()

    def translate(self, sentence, plot=''):
        """
        :param sentence: input sentence you wish to receive a response to
        :param plot: Whether you want to plot the attention weighting
        :return: The response given the input text.
        """

        result, attention_weights = self.evaluate(sentence)

        predicted_sentence = self.tokenizer_tar.decode([i for i in result
                                                        if i < self.tokenizer_tar.vocab_size])

        print('Input: {}'.format(sentence))
        print('Predicted translation: {}'.format(predicted_sentence))

        if plot:
            self.plot_attention_weights(attention_weights, sentence, result, plot)



import tensorflow as tf
from transformer_model.attention import MultiHeadAttention


class EncoderLayer(tf.keras.layers.Layer):

    def __init__(self, d_model, num_heads, dff, rate=0.1):
        """
        :param d_model: The Dimensionality of the Encoder Layer
        :param num_heads: The number of heads to use in the Multiheaded Attention
        :param dff: The input nodes in the initial Dense layer in the Point-wise Feed-Forward NN
        :param rate: The dropout rate to use
        """
        super(EncoderLayer, self).__init__()

        self.mha = MultiHeadAttention(d_model, num_heads)
        self.ffn = MultiHeadAttention.point_wise_feed_forward_network(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, x, training, mask):
        """
        :param x: Input tensor
        :param training: Boolean for whether to train or not.
        :param mask: Dummy masking variable
        :return: Output from the Encoder Layer
        """

        attn_output, _ = self.mha(x, x, x, mask)  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(x + attn_output)  # (batch_size, input_seq_len, d_model)

        ffn_output = self.ffn(out1)  # (batch_size, input_seq_len, d_model)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)  # (batch_size, input_seq_len, d_model)

        return out2



import tensorflow as tf
from transformer_model.encoder import EncoderLayer
from transformer_model.positional_encoder import PositionalEncoder


class Encoder(tf.keras.layers.Layer):

    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size, rate=0.1):
        """
        :param num_layers: Number of EncoderLayers to use
        :param d_model: Dimensionality of the Encoder
        :param num_heads: Number of heads to use in the Multiheaded Attention
        :param dff: input nodes of the Point-wise Feed Forward NN
        :param input_vocab_size: Total number of unique words in the training set
        :param rate: Dropout rate
        """
        super(Encoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(input_vocab_size, d_model)
        self.pos_encoder = PositionalEncoder(d_model=d_model)
        self.pos_encoding = self.pos_encoder.positional_encoding(input_vocab_size, d_model)

        self.enc_layers = [EncoderLayer(d_model, num_heads, dff, rate) for _ in range(num_layers)]

        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, x, training, mask):
        seq_len = tf.shape(x)[1]

        # adding embedding and position encoding.
        x = self.embedding(x)  # (batch_size, input_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, mask)

        return x  # (batch_size, input_seq_len, d_model)




import tensorflow_datasets as tfds
import tensorflow as tf
import pyodbc
import pandas as pd


class DataProcessor(object):

    def __init__(self, test_size=0.1, csv_path=None, sql_source='',
                 max_length=40, feature_col='', target_col='', buffer_size=20000, batch_size=64,
                 server=''):
        """
        :param test_size: The % of data to withhold for the test set
        :param csv_path: The path of the DataFrame to use. If None, queries the sql_source
        :param sql_source: The table to pull the DataFrame from
        :param feature_col: The column name of your feature input. (Defaults to the first column)
        :param target_col:  The column name of your target input. (Defaults ot the second column)
        :param max_length: Maximum number of words per input you want to use for the training set.
        """
        self.test_size = test_size
        self.csv_path = csv_path
        self.sql_source = sql_source
        self.feature_col = feature_col
        self.target_col = target_col
        self.max_length = max_length
        self.buffer_size = buffer_size
        self.batch_size = batch_size

        self.server = server
        self.SQL_SERVER = 'SQL Server'
        self.cnxn = pyodbc.connect(f'DRIVER={self.SQL_SERVER}; SERVER={self.server}; Trusted_Connection=yes')

    def load_data(self):
        """
        :return: Loads and returns the data in a DataFrame.
        """
        if self.csv_path is None:
            df = pd.read_sql(f'SELECT * FROM {self.sql_source}', con=self.cnxn)
            # inst = s.sqlConnect()
            # df = inst.script_to_df(f'SELECT * FROM {self.sql_source}')
            df.dropna(inplace=True)
        else:
            df = pd.read_csv(self.csv_path)

        return df

    def train_test_split(self, df):
        """
        :param df: The DataFrame that you will split into train and test sets.
        :return: Train and Tests DataFrames split on the test_size percentage.
        """
        test_rows = int(len(df) * self.test_size)
        test = df.sample(test_rows)
        train = df[~df.isin(test)]

        train.dropna(inplace=True)
        test.dropna(inplace=True)

        return train, test

    def tokenizer(self, train, vocab_size=2 ** 13):
        """
        :param train: Dataframe containing the training data set.
        :param vocab_size: Max number of words to tokenize
        :return: the Feature tokenizer and Target tokenizer
        """
        if self.feature_col is None:
            feature_col = train.columns[0]
        else:
            feature_col = self.feature_col

        if self.target_col is None:
            target_col = train.columns[1]
        else:
            target_col = self.target_col

        global tokenizer_feat
        global tokenizer_tar

        tokenizer_tar = tfds.features.text.SubwordTextEncoder.build_from_corpus(
            (line for line in train[target_col].values), target_vocab_size=vocab_size
        )

        tokenizer_feat = tfds.features.text.SubwordTextEncoder.build_from_corpus(
            (line for line in train[feature_col].values), target_vocab_size=vocab_size
        )

        return tokenizer_feat, tokenizer_tar

    @staticmethod
    def encode(lang1, lang2):
        """
        :param lang1: A Tensor containing the tokenized client text sentence
        :param lang2: A Tensor containing the tokenized QL TM text sentence
        :return: The originally passed tensors with a Start-of-Sequence (SOS) and
                 a End-of-Sequence (EOS) added.
        """

        lang1 = [tokenizer_feat.vocab_size] + tokenizer_feat.encode(
            lang1.numpy()) + [tokenizer_feat.vocab_size + 1]

        lang2 = [tokenizer_tar.vocab_size] + tokenizer_tar.encode(
            lang2.numpy()) + [tokenizer_tar.vocab_size + 1]

        return lang1, lang2

    def filter_max_length(self, x, y):
        """
        :param x: The Feature tensor
        :param y: The Target tensor
        :return: The pair of Tensors whose lengths are less than or equal to self.max_length
        """

        return tf.logical_and(tf.size(x) <= self.max_length,
                              tf.size(y) <= self.max_length)

    def tf_encode(self, feature, target):
        """
        :param feature: The input feature text
        :param target:  The output target text
        :return: The encoded vector representations of the feature and target text.
        """

        return tf.py_function(self.encode, [feature, target], [tf.int64, tf.int64])

    def to_tensor_dataset(self, data):
        """
        :param data: a DataFrame column that contains the feature_col and target_col
        :return: A TensorDataset made from the data DataFrame.
        """

        return tf.data.Dataset.from_tensor_slices(
            (
                tf.cast(data[self.feature_col].values, tf.string),
                tf.cast(data[self.target_col].values, tf.string)
            )
        )

    def preprocess(self, train: pd.DataFrame, test: pd.DataFrame):
        """

        :param train: Your training set.
        :param test: Your testing set.
        :return: filtered and batched TensorDatasets for train and test.
        """

        train_data = self.to_tensor_dataset(train)
        test_data = self.to_tensor_dataset(test)

        train_dataset = train_data.map(self.tf_encode)
        train_dataset = train_dataset.filter(self.filter_max_length)
        train_dataset = train_dataset.cache()
        train_dataset = train_dataset.shuffle(self.buffer_size).padded_batch(self.batch_size,
                                                                             padded_shapes=([-1], [-1]))
        train_dataset = train_dataset.prefetch(tf.data.experimental.AUTOTUNE)

        test_dataset = test_data.map(self.tf_encode)
        test_dataset = test_dataset.filter(self.filter_max_length).padded_batch(self.batch_size,
                                                                                padded_shapes=([-1], [-1]))

        return train_dataset, test_dataset



import tensorflow as tf
from transformer_model.encoder import Encoder
from transformer_model.decoder import Decoder


class Transformer(tf.keras.Model):
    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size, target_vocab_size, rate=0.1):
        """
        :param num_layers: Number of EncoderLayers and DecoderLayers to use
        :param d_model: Dimensionality of the EncoderLayer and DecoderLayer
        :param num_heads: Number of heads to use in the Multiheaded Attention
        :param dff: Input nodes in the Point-wise Feed Forward NN
        :param input_vocab_size: Total number of unique words in the feature set.
        :param target_vocab_size: Total number of unique words in the target set.
        :param rate: Dropout rate
        """
        super(Transformer, self).__init__()

        self.encoder = Encoder(num_layers, d_model, num_heads, dff,
                               input_vocab_size, rate)

        self.decoder = Decoder(num_layers, d_model, num_heads, dff,
                               target_vocab_size, rate)

        self.final_layer = tf.keras.layers.Dense(target_vocab_size)

    def call(self, inp, tar, training, enc_padding_mask, look_ahead_mask, dec_padding_mask):
        enc_output = self.encoder(inp, training, enc_padding_mask)  # (batch_size, inp_seq_len, d_model)

        # dec_output.shape == (batch_size, tar_seq_len, d_model)
        dec_output, attention_weights = self.decoder(tar, enc_output, training, look_ahead_mask, dec_padding_mask)

        final_output = self.final_layer(dec_output)  # (batch_size, tar_seq_len, target_vocab_size)

        return final_output, attention_weights

    


import tensorflow as tf


class MultiHeadAttention(tf.keras.layers.Layer):

    def __init__(self, d_model, num_heads):
        """
        :param d_model: The Dimensionality of the Attention mechanism
        :param num_heads: The number of heads to use in the attention mechanism
        """
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)

        self.dense = tf.keras.layers.Dense(d_model)

    @staticmethod
    def scaled_dot_product_attention(q, k, v, mask):
        """
        :param q: The query vector
        :param k: The key vector
        :param v: The value vector
        :param mask: Float tensor with shape (..., seq_len_q, seq_len_k). Defaults to None.
        :return: the scaled dot product output and attention weights (output, attention_weights)
        """

        matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)

        # scale matmul_qk
        dk = tf.cast(tf.shape(k)[-1], tf.float32)
        scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

        # add the mask to the scaled tensor.
        if mask is not None:
            scaled_attention_logits += (mask * -1e9)

        # softmax is normalized on the last axis (seq_len_k) so that the scores
        # add up to 1.
        attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)

        output = tf.matmul(attention_weights, v)  # (..., seq_len_q, depth_v)

        return output, attention_weights

    def split_heads(self, x, batch_size):
        """
        :param x: the head input
        :param batch_size:
        :return: splits the last dimension and returns the result with shape (batch_size, num_heads, seq_length, depth)
        """
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        """
        :param v: Value vector
        :param k: Key vector
        :param q: Query vector
        :param mask: Mask shape to use
        :return:
        """
        batch_size = tf.shape(q)[0]

        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)  # (batch_size, seq_len, d_model)
        v = self.wv(v)  # (batch_size, seq_len, d_model)

        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

        # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = self.scaled_dot_product_attention(
            q, k, v, mask)

        scaled_attention = tf.transpose(scaled_attention,
                                        perm=[0, 2, 1, 3])  # (batch_size, seq_len_q, num_heads, depth)

        concat_attention = tf.reshape(scaled_attention,
                                      (batch_size, -1, self.d_model))  # (batch_size, seq_len_q, d_model)

        output = self.dense(concat_attention)  # (batch_size, seq_len_q, d_model)

        return output, attention_weights

    @staticmethod
    def point_wise_feed_forward_network(d_model, dff):
        """
        :param d_model:
        :param dff:
        :return: A feed forward nn to be stacked after each attention layer.
        """

        return tf.keras.Sequential([
            tf.keras.layers.Dense(dff, activation='relu'),  # (batch_size, seq_len, dff)
            tf.keras.layers.Dense(d_model)  # (batch_size, seq_len, d_model)
        ])



import numpy as np
import tensorflow as tf


class PositionalEncoder(object):

    def __init__(self, d_model=512):
        """
        This allows to grade positional importance for each word relative to each sequence it belongs to relative to
        the entire corpus.
        :param d_model: Dimensionality of the encoding.
        """
        self.d_model = d_model

    def get_angles(self, position, i, d_model=None):
        if d_model is None:
            d_model = self.d_model
        angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
        return position * angle_rates

    def positional_encoding(self, position, d_model=None):
        if d_model is None:
            d_model = self.d_model
        angle_rads = self.get_angles(np.arange(position)[:, np.newaxis],
                                     np.arange(d_model)[np.newaxis, :],
                                     d_model)

        # apply sin to even indices in the array; 2i
        sines = np.sin(angle_rads[:, 0::2])

        # apply cos to odd indices in the array; 2i+1
        cosines = np.cos(angle_rads[:, 1::2])

        pos_encoding = np.concatenate([sines, cosines], axis=-1)

        pos_encoding = pos_encoding[np.newaxis, ...]

        return tf.cast(pos_encoding, dtype=tf.float32)



import tensorflow as tf


class Masker(object):

    def __init__(self):
        """
        This class holds a collection of masking functions that are used across the entire package.
        """

    @staticmethod
    def create_padding_mask(seq):
        """
        :param seq: the sequence to mask
        :return: the padding mask in the form of (batch_size, 1, 1, seq_len)
        """
        seq = tf.cast(tf.math.equal(seq, 0), tf.float32)

        # add extra dimensions to add the padding
        # to the attention logits.
        return seq[:, tf.newaxis, tf.newaxis, :]

    @staticmethod
    def create_look_ahead_mask(size):
        """
        :param size:
        :return: the mask for hiding unseen values for training purposes in the form of (seq_len, seq_len)
        """
        mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
        return mask

    def create_masks(self, inp, tar):
        """
        :param self:
        :param inp: The feature tensor to mask.
        :param tar: The target tensor to mask
        :return: the Encoder, Combined, and Decoder masks
        """
        # Encoder padding mask
        enc_padding_mask = self.create_padding_mask(inp)

        # Used in the 2nd attention block in the decoder.
        # This padding mask is used to mask the encoder outputs.
        dec_padding_mask = self.create_padding_mask(inp)

        # Used in the 1st attention block in the decoder.
        # It is used to pad and mask future tokens in the input received by
        # the decoder.
        look_ahead_mask = self.create_look_ahead_mask(tf.shape(tar)[1])
        dec_target_padding_mask = self.create_padding_mask(tar)
        combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

        return enc_padding_mask, combined_mask, dec_padding_mask



from setuptools import setup, find_packages

setup(
    name='transformer_model',
    version='1.1.0',
    packages=find_packages(exclude=['env']),
    url='https://github.com/zbloss/TransformerModel',
    download_url='https://github.com/zbloss/TransformerModel/archive/v1.0.1.tar.gz',
    keywords=['Transformer', 'seq2seq', 'SelfAttention'],
    license='mit',
    author='Zachary Bloss',
    author_email='zacharybloss@gmail.com',
    description='A Transformer model implementation in TensorFlow 2.0',
    python_requires='>=3.5',
    install_requires=[
        'matplotlib',
        'numpy',
        'pandas',
        'pyodbc',
        'tensorflow-datasets>=1.1.0',
        'tensorflow>=2.0.0b1'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
      ],
)
